@@ -52,6 +52,7 @@ from .models.destination import (
     S3Destination, DestinationId, DestinationName
 )
 from .models.destination_reference import TableDestinationReference, FolderDestinationReference
+from .models.event_description import EventDescription, TimestampInfo
 from .models.generated_features import GeneratedFeatures
 from .models.feature import Feature
 from .models.feature_creation_request import EventFeatureCreationRequest, FeatureCreationRequest
@@ -66,7 +67,7 @@ from .models.file_format import CSV, FileFormat, Parquet, Orc
 from .models.jobs import RunStatus, FeatureStoreRunId
 from .models.merge_request import MergeRequest, MergeRequestCreationRequest, MergeRequestId
 from .models.ref import Ref
-from .models.table import Table, TableId
+from .models.table import Table, TableId, TableName
 from .models.aggregate import AggregateExpression
 from .models.attribute import Attribute
 from .models.entity import Entity, EntityId, EntityName
@@ -77,6 +78,9 @@ from .models.label import Label
 from .models.table_preview import NullCell, TablePreview
 from .models.post_aggregate_expression import PostAggregateExpression
 from .models.select_expression import SelectExpression
+from .models.source import Source, SourceId, SourceName
+from .models.source_reference import FolderSourceReference
+from .models.table_creation_request import TableCreationRequest, RootTableCreationRequest, ViewTableCreationRequest
 from . import version
 
 
@@ -382,13 +386,12 @@ class Anaml:
             entity:  Entity Id to filter feature previews on.
             snapshot_date: Date to filter feature preview on.
         """
-        feature_stats = self.preview_feature(feature, entity, snapshot_date)
+        feature_stats = self.preview_feature(feature, entity=entity, snapshot_date=snapshot_date)
         for fs in feature_stats:
-            self._build_feature_plots(fs)
-        self._warn_empty_feature_stats([
-            fs.featureName
-            for fs in feature_stats if isinstance(fs, EmptySummaryStatistics)
-        ])
+            for stats in fs.statistics:
+                if isinstance(stats, EmptySummaryStatistics):
+                    self._warn_empty_feature_stats([fs.featureName])
+                self._build_feature_plots(stats)
 
     def sample_feature(self, feature: Feature) -> pandas.DataFrame:
         """Generate a sample of feature values.
@@ -406,6 +409,33 @@ class Anaml:
 
         return pandas.DataFrame(result)
 
+    def create_feature_set(self, featureset: FeatureSetCreationRequest) -> FeatureSet:
+        """Create a feature set definition on the Anaml server.
+
+        Args:
+            featureSet: The feature Set definition.
+
+        Returns:
+            The new feature Set.
+        """
+        r = self._post("/feature-set", json=featureset.to_json())
+        id = self._int_or_handle_errors(r)
+        return self.get_feature_set_by_id(id)
+
+    def update_feature_set(self, featureset: FeatureSet) -> FeatureSet:
+        """Update an existing feature set definition on the Anaml server.
+
+        Args:
+            featureset: The feature set definition.
+
+        Returns:
+            The feature set definition, with its unique identifier and other computed
+            fields updated.
+        """
+        r = self._put(f"/feature-set/{str(featureset.id.value)}", json=featureset.to_json())
+        _ = self._json_or_handle_errors(r)
+        return self.get_feature_set_by_id(int(featureset.id.value))
+
     def create_feature_template(self, template: FeatureTemplateCreationRequest) -> FeatureTemplate:
         """Create a new feature template.
 
@@ -419,6 +449,33 @@ class Anaml:
         id = self._int_or_handle_errors(r)
         return self.get_feature_template_by_id(id)
 
+    def create_table(self, table: TableCreationRequest) -> Table:
+        """Create a table definition on the Anaml server.
+
+        Args:
+            table: The table Set definition.
+
+        Returns:
+            The new table.
+        """
+        r = self._post("/table", json=table.to_json())
+        id = self._int_or_handle_errors(r)
+        return self.get_table_by_id(id)
+
+    def update_table(self, table: Table) -> Table:
+        """Update an existing table definition on the Anaml server.
+
+        Args:
+            table: The table definition.
+
+        Returns:
+            The table definition, with its unique identifier and other computed
+            fields updated.
+        """
+        r = self._put(f"/table/{str(table.id.value)}", json=table.to_json())
+        _ = self._json_or_handle_errors(r)
+        return self.get_table_by_id(int(table.id.value))
+
     def get_feature_templates(self) -> List[FeatureTemplate]:
         """Get a list of all features templates from the Anaml server.
 
@@ -868,6 +925,32 @@ class Anaml:
         result = self._json_or_handle_errors(r)
         return Entity.from_json(result)
 
+    def get_source_by_id(self, source_id: Union[SourceId, int]) -> Source:
+        """Get a source from the Anaml server.
+
+        Args:
+            source_id: Unique identifier of the source to retrieve.
+
+        Returns:
+            The requested source, if it exists.
+        """
+        r = self._get(f"/source/{int(source_id)}")
+        result = self._json_or_handle_errors(r)
+        return Source.from_json(result)
+
+    def get_source_by_name(self, name: Union[SourceName, str]) -> Source:
+        """Get a source from the Anaml server.
+
+        Args:
+            name: Name of the source to retrieve.
+
+        Returns:
+            The requested source, if it exists.
+        """
+        r = self._get("/source", query={'name': str(name)})
+        result = self._json_or_handle_errors(r)
+        return Source.from_json(result)
+
     # Builder functions
     def build_event_feature(
         self,
@@ -882,6 +965,7 @@ class Anaml:
         attributes: List[Attribute] = [],
         filter: str = None,
         postAggregateExpr: str = None,
+        entityRestrictions: List[EntityId] = None,
         template: TemplateId = None
     ) -> EventFeatureCreationRequest:
         """Helper function for building event features."""
@@ -908,6 +992,7 @@ class Anaml:
             aggregate=aggregate,
             window=window,
             postAggregateExpr=postAggregateExpr,
+            entityRestrictions=entityRestrictions,
             template=template
         )
 
@@ -937,6 +1022,76 @@ class Anaml:
             attributes=attributes,
             features=features)
 
+    def build_external_table(
+        self,
+        *,
+        name: str,
+        description: str,
+        labels: List[Label] = [],
+        attributes: List[Attribute] = [],
+        entities: typing.Dict[EntityId, str] = None,
+        timestampColumn: str = None,
+        timezone: str = None,
+        source: typing.Union[int, str, Source]
+    ) -> RootTableCreationRequest:
+        """Helper function for building external tables."""
+        if entities is not None and timestampColumn is not None:
+            eventDescription = EventDescription(entities, TimestampInfo(timestampColumn, timezone))
+        elif entities is not None and timestampColumn is None:
+            raise ValueError("Please provide value for timestampColumn.")
+        elif entities is None and timestampColumn is not None:
+            raise ValueError("Please provide value for entities.")
+        else:
+            eventDescription = None
+
+        if isinstance(source, Source):
+            source_id = source.id
+        elif isinstance(source, str):
+            source_id = self.get_source_by_name(source).id
+        else:
+            source_id = source
+
+        return RootTableCreationRequest(
+            name=TableName(name),
+            description=description,
+            labels=labels,
+            attributes=attributes,
+            eventDescription=eventDescription,
+            source=FolderSourceReference(source_id))
+
+    def build_view_table(
+        self,
+        *,
+        name: str,
+        description: str,
+        labels: List[Label] = [],
+        attributes: List[Attribute] = [],
+        entities: typing.Dict[EntityId, str] = None,
+        timestampColumn: str = None,
+        timezone: str = None,
+        expression: str,
+        sources: List[TableId] = []
+    ) -> ViewTableCreationRequest:
+        """Helper function for building view tables."""
+        if entities is not None and timestampColumn is not None:
+            eventDescription = EventDescription(entities, TimestampInfo(timestampColumn, timezone))
+        elif entities is not None and timestampColumn is None:
+            raise ValueError("Please provide value for timestampColumn.")
+        elif entities is None and timestampColumn is not None:
+            raise ValueError("Please provide value for entities.")
+        else:
+            eventDescription = None
+
+        return ViewTableCreationRequest(
+            name=TableName(name),
+            description=description,
+            labels=labels,
+            attributes=attributes,
+            eventDescription=eventDescription,
+            expression=expression,
+            sources=sources
+        )
+
     #####################
     # Load feature data #
     #####################
